from typing import TypeVar
import jsonpython


def configclass(filename: str):
    T = TypeVar("T")

    def wrapper(cls: type[T]) -> type[T]:
        def load(self: T) -> None:
            obj = jsonpython.from_file(cls, filename)  # type: ignore
            self.__dict__.update(obj.__dict__)

        cls.__init__ = load

        return cls

    return wrapper


"""
Example config class.

# config.json
# {
#   "lol": 1,
#   "name": "sss"
# }


@config_class("config.json")
class Config:
    lol: int
    name: str

config = Config()
print(config.name) # sss

"""



from .camera import *
from .config import *
from .components import *
from .ui import *
from .scene_manager import *
from .vmath import *
from .game_object import *
from .engine import *
from .events import *
from .draw import *
from .exceptions import *
from .util import *

__all__ = [
    "Camera",
    "Config",
    "UI",
    "Scene",
    "SceneData",
    "Vector2d",
    "GameObject",
    "Engine",
    "Prefab",
    "Component",
    "Transform",
    "Clicked",
    "Border",
    "Text",
    "Label",
    "Button",
    "events",
    "draw",
    "CriticalError",
    "WarningError",
    "NoComponentError",
    "SpriteComponent"
]



from .scene_manager import Scene
from .game_object import GameObject
from .ui import UI
from . import events

import pygame as pg
import logging


class Engine:

    scenes: list[Scene]
    logger: logging.Logger
    scene: Scene
    display: pg.Surface

    def __init__(self, app_name: str, resolution: tuple[int, int] = (800, 600)):
        self.app_name = app_name
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("[%(levelname)s][%(name)s]: %(message)s"))
        self.logger.addHandler(handler)
        self.init(resolution)

    def register(self, scene: Scene):
        self.scenes.append(scene)

    def init(self, resolution: tuple[int, int] = (800, 600)):
        pg.init()
        pg.font.init()
        logging.basicConfig()
        self.display = pg.display.set_mode(resolution)
        pg.display.set_caption(self.app_name)
        logging.info("Game initialized.")

    def run(self):
        try:
            self.cycle()
        except Exception as e:
            logging.critical(e.with_traceback(None))
            exit(1)

    def event_processing(self, event: pg.event.Event):
        if event.type == pg.QUIT:
            logging.info("Quitting")
            exit()

    def iteration(self):
        
        events.set(pg.event.get())
        
        for event in events.get():
            self.event_processing(event)
        
        for i in (GameObject.objects + UI.widgets):
            i.update()

        self.display.fill((0, 0, 0))
        for i in (GameObject.objects + UI.widgets):
            i.draw(self.display)

        pg.display.flip()
        
    def cycle(self):
        while True:
            self.iteration()
            



from ..game_object import GameObject, Prefab
import pygame as pg

class SceneData:
    name: str
    prefabs: list[Prefab]
    
    def __init__(self, name: str):
        self.name = name

class Scene:

    objects: list[GameObject]
    
    def __init__(self, data: SceneData | None = None):
        self.data = data

    def update(self):
        for i in self.objects:
            i.update()
    
    def draw(self, display: pg.Surface):
        for i in self.objects:
            i.draw(display)



from typing import TypeVar

T = TypeVar("T")


class Seq:
    """
    Customized container. TODO
    """
    __cont: list[T]

    def __init__(self):
        __cont = []

    def get_by_name(self):
        pass



import pygame as pg

from .camera import camera

def text(display: pg.Surface, 
         text: str, 
         pos: tuple[int, int], 
         font: pg.font.Font, 
         color: tuple[int, int, int], 
         background: tuple[int, int, int] = None):
    text_surface = font.render(text, True, color, background)
    new_pos = camera.normalize((pos[0], pos[1])).as_tuple()
    text_rect = (new_pos[0], new_pos[1], text_surface.get_width(), text_surface.get_height())

    display.blit(text_surface, text_rect)

def rect(display: pg.Surface, rect: pg.Rect, color: tuple[int, int, int], width: int = 0):
    new_pos = camera.normalize((rect.x, rect.y)).as_tuple()
    rect = (new_pos[0], new_pos[1], rect.width, rect.height)
    pg.draw.rect(display, color, rect, width)

def circle(display: pg.Surface, pos: tuple[int, int], radius: int, color: tuple[int, int, int], width: int = 0):
    new_pos = camera.normalize(pos)
    pg.draw.circle(display, color, new_pos, radius, width)

def ellipse(display: pg.Surface, rect: pg.Rect, color: tuple[int, int, int], width: int = 0):
    new_pos = camera.normalize((rect.x, rect.y)).as_tuple()
    rect = (new_pos[0], new_pos[1], rect.width, rect.height)
    pg.draw.ellipse(display, color, rect, width)

def line(display: pg.Surface, start_pos: tuple[int, int], end_pos: tuple[int, int], color: tuple[int, int, int], width: int = 1):
    new_start_pos = camera.normalize(start_pos)
    new_end_pos = camera.normalize(end_pos)
    pg.draw.line(display, color, new_start_pos, new_end_pos, width)



import pygame
from .. import events

class Clicked:
    def __init__(self, color: tuple[int, int, int], border: "Border | None" = None):
        self.color = color
        self.border = border

    @staticmethod
    def get_default():
        return Clicked((0, 0, 0))


class Border:

    def __init__(self, width: int, color: tuple[int, int, int], radius: int):
        self.width = width
        self.color = color
        self.radius = radius

    @staticmethod
    def get_default():
        return Border(1, (0, 0, 0), -1)


class Text:

    def __init__(self, text: str,
                 font: pygame.font.Font | None = None,
                 color: tuple[int, int, int] = (0, 0, 0),
                 background_color: tuple[int, int, int] | None = None):
        font = font or pygame.font.SysFont("Arial", 20)
        self.text = text
        self.font = font
        self.color = color
        self.background_color = background_color

    def render(self):
        return self.font.render(self.text, False, self.color, self.background_color)


class RawWidget:
    """
    Abstract widget class.
    """

    def __init__(self, x: int, y: int, width: int = 100, height: int = 100) -> None:
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.childs: list[RawWidget] = []

    def draw(self, window: pygame.Surface):
        for i in self.childs:
            i.draw(window)

    def update(self):
        for i in self.childs:
            i.update()


class Widget(RawWidget):

    def __init__(self, x: int, y: int, width: int, height: int,
                 clicked: Clicked | None = None, color=None, border: Border | None = None,
                 text: Text | None = None):
        super().__init__(x, y, width, height)
        self.text = text
        self.border = border
        self.color = color
        self.clicked = clicked
        self.is_clicked = False
        UI.add_widget(self)

    def draw(self, window: pygame.Surface):
        super().draw(window)

        color = self.color
        border = self.border
        if self.clicked and self.is_clicked:
            color = self.clicked.color or color
            border = self.clicked.border or border

        if border:
            if color:
                pygame.draw.rect(window, color, (self.x, self.y, self.width, self.height), 0, border.radius)
            bw = border.width
            pygame.draw.rect(window, border.color, (self.x - bw, self.y - bw, self.width + bw, self.height + bw), bw,
                             border.radius)
        elif color:
            pygame.draw.rect(window, color, (self.x, self.y, self.width, self.height))

        if self.text:
            txt = self.text.font.render(self.text.text, False, self.text.color)
            window.blit(txt, (self.x + (self.width - txt.get_width()) / 2,
                              self.y + (self.height - txt.get_height()) / 2) + txt.get_size())

    def update(self):
        super().update()
        evnts = events.get()

        rc = pygame.Rect(self.x, self.y, self.width, self.height)
        for i in evnts:
            if i.type == pygame.MOUSEBUTTONDOWN and rc.collidepoint(pygame.mouse.get_pos()):
                self.is_clicked = True
            if i.type == pygame.MOUSEBUTTONUP and self.is_clicked:
                self.is_clicked = False
                self.onclick()

    def onclick(self):
        pass


class Label(Widget):
    """
    Shorthand name of widget that is used only as text.
    """

    def __init__(self, x, y, text: Text | None = None):
        self.text = text or Text('')
        width, height = self.text.render().get_size()

        super().__init__(x, y, width, height, text=text)

    def draw(self, window: pygame.Surface):
        super().draw(window)

    def update(self, events):
        super().update(events)


class Button(Widget):

    def __init__(self, x: int, y: int, width: int, height: int,
                 clicked: Clicked | None = None, color=(255, 0, 0), border: Border | None = None, cmd=lambda _: (),
                 text: Text | None = None, data=None):
        Widget.__init__(self, x, y, width, height, clicked, color, border, text)
        self.cmd = cmd
        self.data = data

    def onclick(self):
        self.cmd(self.data)


class UI:
    widgets: list[Widget] = []

    name: str
    is_disabled: bool

    def __init__(self, name: str) -> None:
        self.name = name
        self.is_disabled = False

    @staticmethod
    def add_widget(widget: Widget):
        UI.widgets.append(widget)

    def draw(self, window: pygame.Surface):
        if self.is_disabled: return
        for i in self.widgets:
            i.draw(window)

    def update(self, events: list[pygame.event.Event]):
        if self.is_disabled: return
        for i in self.widgets:
            i.update(events)
